import numpy as np
import cv2

sub_count = 0


def almost_equal(a, b):
    if abs(a-b) < pow(10, -4):
        return True
    else:
        return False


def reflect_point(point, k, n):
    # print(point)
    x, y = point[0], point[1]
    if almost_equal(0, k):
        return (x, -y + 2*n)
    # calculate perpendicular line
    k_p = -1/k
    n_p = y - k_p * x
    # finding x coordinate of the interception
    x_mid = (n_p - n) / (k - k_p)
    y_mid = k * x_mid + n
    # vector connecting point to midpoint
    PM_vect = np.array([x_mid, y_mid]) - np.array([x, y])
    # reflection point
    R_vect = np.array([x, y]) + 2*PM_vect
    # converting to tupple
    return tuple(R_vect)


def get_mass_center(cnt):
    M = cv2.moments(cnt)
    epsilon = 1e-7
    cx = int(M["m10"] / (M["m00"] + epsilon))
    cy = int(M["m01"] / (M["m00"] + epsilon))
    return (cx, cy)


def get_side(cnt, k, n):
    center = get_mass_center(cnt)
    # points on the line
    x1, y1 = 0, n
    x2, y2 = center[0], k * center[0] + n
    # center of mass
    x_m, y_m = center
    # parallelogram surface area
    cross_product = ((x2 - x1)*(y_m - y1) - (y2 - y1)*(x_m - x1))
    # negative if vectors point in different directions
    return cross_product < 0  # return True if on right side, False otherwise


def overlap(cnt1, cnt2, img_shape):
    # drawing each contour on empty images
    img1 = np.zeros(img_shape, dtype=np.uint8)
    img2 = np.zeros(img_shape, dtype=np.uint8)
    cv2.drawContours(img1, [cnt1], -1, (255, 255, 255), -1)
    cv2.drawContours(img2, [cnt2], -1, (255, 255, 255), -1)

    # comparing overlap to union
    overlap = cv2.bitwise_and(img1, img2)
    overlap_area = cv2.countNonZero(overlap)
    cnt1_area = cv2.countNonZero(img1)
    cnt2_area = cv2.countNonZero(img2)
    union = cnt1_area + cnt2_area - overlap_area

    # returning percentage
    return overlap_area / union * 100


def test_line(cnt_list, line, img_shape, tolerance, overlap_thresh):
    # Extract slope and intercept from line
    slope, intercept = line

    # Initialize matched indices, overlap percentages and matched contours
    overlap_percents = []
    matched_contours = []

    # Separate contours into right and left side based on their center of mass
    right_side_cnts = [
        cnt for cnt in cnt_list if get_side(cnt, slope, intercept)]
    left_side_cnts = [
        cnt for cnt in cnt_list if not get_side(cnt, slope, intercept)]

    # 1. Reflect right side contours and check for potential matches on the left
    for right_cnt in right_side_cnts:
        # Reflect contour center
        reflected_center = reflect_point(
            get_mass_center(right_cnt), slope, intercept)

        for left_cnt in left_side_cnts:
            left_center = get_mass_center(left_cnt)

            # If the reflected center is close to a left side center
            if np.linalg.norm(np.array(reflected_center) - np.array(left_center)) < tolerance:
                # Reflect the entire contour and calculate overlap
                reflected_cnt = [reflect_point(
                    point[0], slope, intercept) for point in right_cnt]
                # reshape into contour format
                reflected_cnt = np.array(reflected_cnt).reshape(
                    (-1, 1, 2)).astype(np.int32)
                left_cnt = np.array(left_cnt).reshape(
                    (-1, 1, 2)).astype(np.int32)
                overlap_percent = overlap(
                    reflected_cnt, left_cnt, img_shape[:2])

                # If overlap is above the threshold, save the match
                if overlap_percent >= overlap_thresh:
                    overlap_percents.append(overlap_percent)
                    matched_contours.append([left_cnt, right_cnt])

    # 2. Check points close to the line
    for cnt in cnt_list:
        # Only proceed if contour has not been matched already
        if not any(cnt is pair_item for pair in matched_contours for pair_item in pair):
            center = get_mass_center(cnt)

            # If the center is close to the line
            if abs(center[1] - (slope * center[0] + intercept)) < tolerance:
                # Reflect the contour and calculate overlap with itself
                reflected_cnt = [reflect_point(
                    point[0], slope, intercept) for point in cnt]
                # reshape into contour format
                cnt = np.array(cnt).reshape((-1, 1, 2)).astype(np.int32)
                reflected_cnt = np.array(reflected_cnt).reshape(
                    (-1, 1, 2)).astype(np.int32)

                overlap_percent = overlap(reflected_cnt, cnt, img_shape[:2])

                # If overlap is above the threshold, save the match
                if overlap_percent >= overlap_thresh:
                    overlap_percents.append(overlap_percent)
                    matched_contours.append([cnt, cnt])

    avg_overlap_percent = sum(overlap_percents) / \
        len(overlap_percents) if overlap_percents else 0

    return matched_contours, avg_overlap_percent


def draw_line_short(image, point, angle, length, color=(255, 0, 0), thickness=1):
    # draw the point of interest in red
    cv2.circle(image, point, 3, (0, 0, 255), -1)
    # calculate the endpoints of the line
    x1 = int(point[0] - length * np.cos(angle))
    y1 = int(point[1] - length * np.sin(angle))
    end_point1 = (x1, y1)

    x2 = int(point[0] + length * np.cos(angle))
    y2 = int(point[1] + length * np.sin(angle))
    end_point2 = (x2, y2)

    cv2.line(image, end_point1, end_point2, color, thickness)


def draw_reflections(cnt_list, line, img_shape, tolerance):
    # Extract slope and intercept from line
    slope, intercept = line
    reflected_cnts = []

    # Separate contours into right and left side based on their center of mass
    left_side_cnts = [
        cnt for cnt in cnt_list if get_side(cnt, slope, intercept)]
    right_side_cnts = [
        cnt for cnt in cnt_list if not get_side(cnt, slope, intercept)]

    # 1. Reflect right side contours and check for potential matches on the left
    for right_cnt in right_side_cnts:
        # Reflect contour center
        reflected_center = reflect_point(
            get_mass_center(right_cnt), slope, intercept)

        for left_cnt in left_side_cnts:
            left_center = get_mass_center(left_cnt)

            # If the reflected center is close to a left side center
            if np.linalg.norm(np.array(reflected_center) - np.array(left_center)) < tolerance:
                # Reflect the entire contour and calculate overlap
                reflected_cnt = [reflect_point(
                    point[0], slope, intercept) for point in right_cnt]
                # reshape into contour format
                reflected_cnt = np.array(reflected_cnt).reshape(
                    (-1, 1, 2)).astype(np.int32)
                reflected_cnts.append(reflected_cnt)

    # draw the line
    image = np.full((img_shape[0], img_shape[1], 3), 255, np.uint8)

    # original contours
    cv2.drawContours(image, cnt_list, -1, (0, 0, 0), 2)
    # line
    slope, intercept = line
    x1, y1 = 0, int(intercept)
    x2, y2 = image.shape[1], int(slope * image.shape[1] + intercept)
    cv2.line(image, (x1, y1), (x2, y2), (0, 0, 255), 2)

    # display reflected contours
    cv2.drawContours(image, reflected_cnts, -1, (0, 0, 255), 2)
    cv2.imshow("Reflected contours", image)
    cv2.waitKey(0)


def get_best_symmetry_line(cnt_list, img_shape, step=50, angle_step=5, tolerance=20, overlap_thresh=80):
    global sub_count, point_count
    hull = cv2.convexHull(np.concatenate(cnt_list))

    hull_image = np.full((img_shape[0], img_shape[1], 3), 255, np.uint8)
    cv2.drawContours(hull_image, cnt_list, -1, (0, 0, 0), 2)

    for i in range(len(hull)):
        cv2.line(hull_image, tuple(hull[i][0]), tuple(
            hull[(i+1) % len(hull)][0]), (0, 0, 255), 1)

    # Compute the total length of the convex hull
    total_len = 0.0
    for i in range(len(hull)):
        total_len += np.linalg.norm(hull[i][0] - hull[(i+1) % len(hull)][0])

    # Determine the number of points to distribute
    num_points = int(np.round(total_len / step))

    # Distribute points evenly along the hull
    event_points = []
    current_dist = 0.0  # current accumulated distance along the hull
    current_point_idx = 0  # index of the current point on the hull
    for _ in range(num_points):
        # While the distance to the next point on the hull is less than the current accumulated distance plus the step,
        # move to the next point on the hull
        while np.linalg.norm(hull[current_point_idx][0] - hull[(current_point_idx+1) % len(hull)][0]) < current_dist + step:
            current_dist -= np.linalg.norm(hull[current_point_idx]
                                           [0] - hull[(current_point_idx+1) % len(hull)][0])
            current_point_idx = (current_point_idx + 1) % len(hull)
        # Interpolate to get the location of the current point
        t = (current_dist + step) / np.linalg.norm(
            hull[current_point_idx][0] - hull[(current_point_idx+1) % len(hull)][0])
        current_point = ((1-t) * hull[current_point_idx][0] + t *
                         hull[(current_point_idx+1) % len(hull)][0]).astype(int)
        event_points.append(tuple(current_point))
        current_dist += step

    # for displaying the hull with points
    for point in event_points:
        cv2.circle(hull_image, point, 3, (0, 0, 255), -1)

    # display the image with convex hull
    # cv2.imshow("Hull image", hull_image)
    # cv2.waitKey(0)
    # cv2.imwrite(f'results/subs/sub{sub_count}.png', hull_image)

    angles = [np.radians(angle) for angle in range(
        0, 180, angle_step)]

    best_matches = []
    best_overlap = 0
    best_line = None

    for event_point in event_points:

        bundle_image = np.copy(hull_image)
        cv2.drawContours(bundle_image, cnt_list, -1, (0, 0, 0), 2)

        for angle in angles:
            draw_line_short(bundle_image, event_point, angle, 1500)
            slope = np.tan(angle)
            intercept = event_point[1] - slope * event_point[0]
            matches, avg_overlap = test_line(
                cnt_list, (slope, intercept), img_shape, tolerance, overlap_thresh)
            if len(matches) > len(best_matches) or (len(matches) == len(best_matches) and avg_overlap > best_overlap):
                best_matches = matches
                best_overlap = avg_overlap
                best_line = (slope, intercept)

    # display reflected contours
    # draw_reflections(cnt_list, best_line, img_shape, tolerance)

    return best_line, best_matches, best_overlap
